package stamp.util.history;
import stamp.util.text.*;

/**
 * Abstract class for storing history messages into a byte array.
 *
 * The byte array has following layout:
 *  addr = 0     DW size       2 bytes, size of table
 *  addr = 2     DW first      2 bytes, address of first header in message area (to skip wrap around bytes)
 *  addr = 4     DS msgsize    start of message area, wraps around to 4
 *               ....          msgsize = size - 4
 *  addr = size                table end, message area end
 *
 * Each message has following layout:
 *  DB len     messages can only be 255 bytes maximum, len=0 identifies new entry
 *  DS len     message (len bytes), address (1 word) of oldest message in case new entry
 *
 * Total size = 4 + msgsize = 4 + NumberOfMessages*(1 + AverageSizeMessages) + 3
 * Example: to support 10 messages with average length 32 bytes
 * size = 4 + 10*(1+32) + 3 = 337
 *
 * Example of table
 *
 *               +-------------+
 *  0x0000       |  size_lsb   |
 *               |  size_msb   |
 *               +-------------+
 *    +----------| first_lsb   |
 *    |          | first_msb   |
 *    |          +-------------+
 *    |         ///  message  /// wrap around
 *    |          +-------------+
 *  first        |     len     |  first header
 *              ///  message  ///
 *               +-------------+
 *               :    ....     :
 *               +-------------+
 *               |     len     |  latest message
 *              ///  message  ///
 *               +-------------+
 *               |   len = 0   |   new entry
 *               | oldest_lsb  |------+  oldest=0 if no messages
 *               | oldest_msb  |      |
 *              ///  unused   ///     |
 *               +-------------+      |
 *               |     len     |   oldest message
 *              ///  message  ///
 *               +-------------+
 *               |     len     |
 *              ///  message  /// wrap around
 *               +-------------+
 *  size
 *
 * @version 1.0, November 17, 2008
 * @author Peter Verkaik (verkaik6@zonnet.nl)
 */

public abstract class History {

  protected abstract void writeByte(int addr, int val); //write byte to storage area
  protected abstract void writeWord(int addr, int val); //write word to storage area
  protected abstract int readByte(int addr);            //read byte from storage area
  protected abstract int readWord(int addr);            //read word from storage area

  private static final int MSGAREA = 0x0004;            //start of message area

  private int size;           //size of table
  private int first;          //address of first header (any message)
  private int oldestMessage;  //address of oldest message
  private int latestMessage;  //address of latest message
  private int newMessage;     //address to store new message
  private int maxLen;         //maximum length of message
  private int readPtr;

  /**
   * Constructor.
   *
   * @param size Restores or clears an array of size bytes for message storage.
   */
  public History(int size) {
    this.size = size;
    maxLen = size - MSGAREA - 4; //the 4 represent len byte plus new entry bytes
    if (maxLen > 255) maxLen = 255; //messages are truncated to maxLen
  }

  /**
   * Initialize.
   * This method must be called from the extended class constructor after super().
   */
  protected void init() {
    if (size == readWord(0)) restore();
    else clear();
  }

  /**
   * Clear message array.
   * This method removes all present messages.
   */
  public void clear() {
    first = MSGAREA;
    oldestMessage = 0;
    latestMessage = 0;
    newMessage = first;
    writeWord(0,size);
    writeWord(2,first);
    for (int addr=MSGAREA; addr<size; addr++) writeByte(addr,0);
  }

  /**
   * Restore message array.
   * This method restores the message array, that is it sets the
   * running parameters to values matching the message array.
   */
  private void restore() {
    int addr,len;
    first = readWord(2);
    oldestMessage = 0;
    latestMessage = 0;
    newMessage = 0;
    addr = first;
    do {
      len = readByte(addr);     //get len
      if (len == 0) {
        newMessage = addr++;
        if (addr == size) addr = MSGAREA;  //in case wrap around
        oldestMessage = readByte(addr++);
        if (addr == size) addr = MSGAREA;
        oldestMessage |= (readByte(addr)<<8);
        addr = size; //while exits here
      }
      else {
        latestMessage = addr;
        addr += len+1;  //move to next message, no need to check for wrap around
      }
    } while (addr < size);
    if ((latestMessage == 0) && (oldestMessage != 0)) { //find latestMessage, this is only when newMessage == 4 + offset
      addr = oldestMessage;
      do {
        len = readByte(addr);     //get len
        if (len == 0) {
          addr = size; //while exits here
        }
        else {
          latestMessage = addr;
          addr += len+1;  //move to next message
          if (addr >= size) addr -= (size-MSGAREA); //in case wrap around
        }
      } while (addr < size);
    }
  }

/*
  When writing a new message it is possible to overwrite multiple older messages,
  the first being the oldest message. The oldest message may be wrapped around,
  or this new message may wrap around. Therefore the first header may change as well.
*/

  /**
   * Write message.
   * This method resets the read pointer, so after calling this method,
   * reading messages should start with readOldest()
   *
   * @param mesg Array to be written
   * @param num Maximum number of bytes to write
   * @param offset Start index in mesg
   * @return Actual number of bytes written (0 if message not written)
   */
  public int write(char[] mesg, int num, int offset) {
    int addr,len,newFirst,unused;
    if (num > 0) {
      readPtr = 0; //reset read pointer
      newFirst = 0;
      if (num > maxLen) num = maxLen; //truncate message if necessary
      latestMessage = newMessage; //this new message will be latest message
      //check if this new message + new entry overwrites older messages
      addr = newMessage;
      do {
        if (oldestMessage != 0) {
          unused = oldestMessage - addr;
          if (unused < 0) unused += (size-MSGAREA);
          if (unused < (num + 4)) { //4 is len byte plus new entry bytes, new message does not fit unused bytes
            oldestMessage += readByte(oldestMessage) + 1; //1 is for len byte
            if (oldestMessage >= size) { //oldestMessage wrapped around
              oldestMessage -= (size-MSGAREA);
            }
            if (oldestMessage == addr) {
              oldestMessage = 0;
            }
          }
        }
        else {
          oldestMessage = latestMessage;
          break;
        }
      } while (unused < (num+4));
      //calculate new first
      newFirst = addr + num + 1;
      if (newFirst >= size) { //wraps around, new entry after this new message will be new first
        newFirst -= (size - MSGAREA);
        first = newFirst;
        writeWord(2,first);
      }
      else {
        if (oldestMessage < newFirst) {
          first = oldestMessage;
          writeWord(2,first);
        }
      }
      len = num;
      //write header
      writeByte(newMessage++,len); //len=num
      if (newMessage == size) { //this new message wrappes around
        newMessage = MSGAREA;
      }
      //write message
      while (len-- > 0) {
        writeByte(newMessage++,mesg[offset++]);
        if (newMessage == size) { //this new message wrappes around
          newMessage = MSGAREA;
        }
      }
      //write new entry
      addr = newMessage;
      writeByte(addr++,0); //len=0
      if (addr == size) {
        addr = MSGAREA;
      }
      writeByte(addr++,oldestMessage);
      if (addr == size) {
        addr = MSGAREA;
      }
      writeByte(addr,(oldestMessage>>>8));
      return num;
    }
    return 0;
  }

  /**
   * Read oldest message.
   *
   * @param mesg Array to be filled with message
   * @param num Maximum number of bytes to read
   * @param offset Start index in mesg
   * @return Actual number of bytes read (0 if no messages)
   */
  public int readOldest(char[] mesg, int num, int offset) {
    int i,c,addr,len;
    readPtr = oldestMessage;
    addr = readPtr;
    if (addr != 0) {
      len = readByte(addr++);
      if (addr == size) addr = MSGAREA;
      if (len < num) num = len;
      for (i=0; i<num; i++) {
        c = readByte(addr++);
        if (addr == size) addr = MSGAREA;
        mesg[offset++] = (char)c;
      }
      return num;
    }
    return 0;
  }

  /**
   * Read latest message.
   *
   * @param mesg Array to be filled with message
   * @param num Maximum number of bytes to read
   * @param offset Start index in mesg
   * @return Actual number of bytes read (0 if no messages)
   */
  public int readLatest(char[] mesg, int num, int offset) {
    int i,c,addr,len;
    readPtr = latestMessage;
    addr = readPtr;
    if (addr != 0) {
      len = readByte(addr++);
      if (addr == size) addr = MSGAREA;
      if (len < num) num = len;
      for (i=0; i<num; i++) {
        c = readByte(addr++);
        if (addr == size) addr = MSGAREA;
        mesg[offset++] = (char)c;
      }
      return num;
    }
    return 0;
  }

  /**
   * Read next message.
   *
   * @param mesg Array to be filled with message
   * @param num Maximum number of bytes to read
   * @param offset Start index in mesg
   * @return Actual number of bytes read (0 if no more messages)
   */
  public int readNext(char[] mesg, int num, int offset) {
    int i,c,addr,len;
    addr = readPtr;
    if (addr != 0) {
      len = readByte(addr);
      if (len != 0) {
        addr += len+1;
        if (addr >= size) addr -= (size-MSGAREA);
        readPtr = addr;
        len = readByte(addr++);
        if (addr == size) addr = MSGAREA;
        if (len < num) num = len;
        for (i=0; i<num; i++) {
          c = readByte(addr++);
          if (addr == size) addr = MSGAREA;
          mesg[offset++] = (char)c;
        }
        return num;
      }
    }
    return 0;
  }

  /**
   * Print message table to IDE message window as HEX numbers.
   */
  public void printTable() {
    int x,y;
    char c;
    x = 0;
    int num = size;
    while (num > 0) {
      Format.printf("%04x : ",(x*16));   // Display the address
      int len = (num < 16) ? num : 16;
      for (y=0;y<len;y++) {              // write the contents in hex
        Format.printf("%02x ",readByte((x*16)+y));
      }
      for (y=0;y<len;y++) {              // Display the ASCII-characters
        c = (char)readByte((x*16)+y);
        if ((c>=' ') && (c<=0x7E))
          Format.printf("%c",c);
        else
          Format.printf(".");
      }
      Format.printf("\n");               // New line
      num -= len;                        // adjust for remaining characters
      x++;                               // increment for address
    }
  }

  /**
   * Print character array to IDE message window as HEX numbers.
   */
  public void printArray(char[] txt, int num, int offset) {
    int x,y;
    char c;
    x = 0;
    while (num > 0) {
      Format.printf("%04x : ",(x*16));   // Display the address
      int len = (num < 16) ? num : 16;
      for (y=0;y<len;y++) {              // write the contents in hex
        Format.printf("%02x ",0xFF & txt[(x*16)+y+offset]);
      }
      for (y=0;y<len;y++) {              // Display the ASCII-characters
        c = txt[(x*16)+y+offset];
        if ((c>=' ') && (c<=0x7E))
          Format.printf("%c",c);
        else
          Format.printf(".");
      }
      Format.printf("\n");               // New line
      num -= len;                        // adjust for remaining characters
      x++;                               // increment for address
    }
  }

}